@@ -1,20 +1,24 @@
-from transitions.extensions import MachineFactory
+from transitions.extensions.asyncio import AsyncMachine, HierarchicalAsyncMachine
+from transitions.extensions.factory import AsyncGraphMachine, HierarchicalAsyncGraphMachine
 
 try:
     import asyncio
 except (ImportError, SyntaxError):
-    asyncio = None
+    asyncio = None  # type: ignore
 
 
 from unittest.mock import MagicMock
 from unittest import skipIf
 from functools import partial
 import weakref
-from .test_core import TestTransitions, MachineError
+from .test_core import TestTransitions, MachineError, TYPE_CHECKING
 from .utils import DummyModel
 from .test_graphviz import pgv as gv
 from .test_pygraphviz import pgv
 
+if TYPE_CHECKING:
+    from typing import Type
+
 
 @skipIf(asyncio is None, "AsyncMachine requires asyncio and contextvars suppport")
 class TestAsync(TestTransitions):
@@ -49,7 +53,7 @@ class TestAsync(TestTransitions):
 
     def setUp(self):
         super(TestAsync, self).setUp()
-        self.machine_cls = MachineFactory.get_predefined(asyncio=True)
+        self.machine_cls = AsyncMachine  # type: Type[AsyncMachine]
         self.machine = self.machine_cls(states=['A', 'B', 'C'], transitions=[['go', 'A', 'B']], initial='A')
 
     def test_new_state_in_enter_callback(self):
@@ -87,7 +91,7 @@ class TestAsync(TestTransitions):
             mock()
 
         m = self.machine
-        m.after_state_change = async_process
+        m.after_state_change = [async_process]
         asyncio.run(m.go())
         self.assertEqual(m.state, 'B')
         self.assertTrue(mock.called)
@@ -308,6 +312,7 @@ class TestAsync(TestTransitions):
         async def run():
             m1 = DummyModel()
             m2 = DummyModel()
+            self.machine_cls = HierarchicalAsyncMachine
             m = self.machine_cls(model=[m1, m2], states=['A', 'B', 'C'], transitions=transitions,
                                  initial='A', queued=True, send_event=True)
             await asyncio.gather(m1.go(), m2.go(),
@@ -336,10 +341,12 @@ class TestAsync(TestTransitions):
         timeout_called = MagicMock()
 
         @add_state_features(AsyncTimeout)
-        class TimeoutMachine(self.machine_cls):
+        class TimeoutMachine(self.machine_cls):  # type: ignore
             pass
 
-        states = ['A', {'name': 'B', 'timeout': 0.2, 'on_timeout': ['to_C', timeout_called]}, 'C']
+        states = ['A',
+                  {'name': 'B', 'timeout': 0.2, 'on_timeout': ['to_C', timeout_called]},
+                  {'name': 'C', 'timeout': 0, 'on_timeout': 'to_D'}, 'D']
         m = TimeoutMachine(states=states, initial='A')
         with self.assertRaises(AttributeError):
             m.add_state('Fail', timeout=1)
@@ -354,6 +361,12 @@ class TestAsync(TestTransitions):
             await asyncio.sleep(0.3)
             self.assertTrue(m.is_C())  # now timeout should have been processed
             self.assertTrue(timeout_called.called)
+            m.get_state('C').timeout = 0.05
+            await m.to_B()
+            await asyncio.sleep(0.3)
+            self.assertTrue(m.is_D())
+            self.assertEqual(2, timeout_called.call_count)
+
         asyncio.run(run())
 
     def test_callback_order(self):
@@ -403,10 +416,10 @@ class TestAsync(TestTransitions):
         mock = MagicMock()
 
         def on_exception(event_data):
-            self.assertIsInstance(event_data.error, ValueError)
+            self.assertIsInstance(event_data.error, (ValueError, MachineError))
             mock()
 
-        m = self.machine_cls(states=['A', 'B'], initial='A', send_event=True,
+        m = self.machine_cls(states=['A', 'B'], initial='A', transitions=[['go', 'A', 'B']], send_event=True,
                              after_state_change=partial(self.stuff.this_raises, ValueError))
 
         async def run():
@@ -415,6 +428,9 @@ class TestAsync(TestTransitions):
 
             m.on_exception.append(on_exception)
             await m.to_B()
+            await m.go()
+            self.assertTrue(mock.called)
+            self.assertEqual(2, mock.call_count)
             self.assertTrue(mock.called)
 
         asyncio.run(run())
@@ -426,13 +442,54 @@ class TestAsync(TestTransitions):
         asyncio.run(pr.go())
         self.assertTrue(pr.is_B())
 
+    def test_may_transition_with_auto_transitions(self):
+        states = ['A', 'B', 'C']
+        d = DummyModel()
+        self.machine_cls(model=d, states=states, initial='A')
+
+        async def run():
+            assert await d.may_to_A()
+            assert await d.may_to_B()
+            assert await d.may_to_C()
+
+        asyncio.run(run())
+
+    def test_machine_may_transitions(self):
+        states = ['A', 'B', 'C']
+        m = self.machine_cls(states=states, initial='A', auto_transitions=False)
+        m.add_transition('walk', 'A', 'B', conditions=[lambda: False])
+        m.add_transition('stop', 'B', 'C')
+        m.add_transition('run', 'A', 'C')
+
+        async def run():
+            assert not await m.may_walk()
+            assert not await m.may_stop()
+            assert await m.may_run()
+            await m.run()
+            assert not await m.may_run()
+            assert not await m.may_stop()
+            assert not await m.may_walk()
+
+        asyncio.run(run())
+
+    def test_may_transition_with_invalid_state(self):
+        states = ['A', 'B', 'C']
+        d = DummyModel()
+        m = self.machine_cls(model=d, states=states, initial='A', auto_transitions=False)
+        m.add_transition('walk', 'A', 'UNKNOWN')
+
+        async def run():
+            assert not await d.may_walk()
+
+        asyncio.run(run())
+
 
 @skipIf(asyncio is None or (pgv is None and gv is None), "AsyncGraphMachine requires asyncio and (py)gaphviz")
-class AsyncGraphMachine(TestAsync):
+class TestAsyncGraphMachine(TestAsync):
 
     def setUp(self):
         super(TestAsync, self).setUp()
-        self.machine_cls = MachineFactory.get_predefined(graph=True, asyncio=True)
+        self.machine_cls = AsyncGraphMachine  # type: Type[AsyncGraphMachine]
         self.machine = self.machine_cls(states=['A', 'B', 'C'], transitions=[['go', 'A', 'B']], initial='A')
 
 
@@ -440,7 +497,7 @@ class TestHierarchicalAsync(TestAsync):
 
     def setUp(self):
         super(TestAsync, self).setUp()
-        self.machine_cls = MachineFactory.get_predefined(nested=True, asyncio=True)
+        self.machine_cls = HierarchicalAsyncMachine  # type: Type[HierarchicalAsyncMachine]
         self.machine = self.machine_cls(states=['A', 'B', 'C'], transitions=[['go', 'A', 'B']], initial='A')
 
     def test_nested_async(self):
@@ -475,9 +532,9 @@ class TestHierarchicalAsync(TestAsync):
 
 
 @skipIf(asyncio is None or (pgv is None and gv is None), "AsyncGraphMachine requires asyncio and (py)gaphviz")
-class AsyncHierarchicalGraphMachine(TestHierarchicalAsync):
+class TestAsyncHierarchicalGraphMachine(TestHierarchicalAsync):
 
     def setUp(self):
         super(TestHierarchicalAsync, self).setUp()
-        self.machine_cls = MachineFactory.get_predefined(graph=True, asyncio=True, nested=True)
+        self.machine_cls = HierarchicalAsyncGraphMachine  # type: Type[HierarchicalAsyncGraphMachine]
         self.machine = self.machine_cls(states=['A', 'B', 'C'], transitions=[['go', 'A', 'B']], initial='A')
