@@ -4,20 +4,25 @@ except ImportError:
     pass
 
 import sys
+from typing import TYPE_CHECKING
+from functools import partial
+from unittest import TestCase, skipIf
+import weakref
+
+from transitions import Machine, MachineError, State, EventData
+from transitions.core import listify, _prep_ordered_arg
 
 from .utils import InheritedStuff
 from .utils import Stuff, DummyModel
 
-from functools import partial
-from transitions import Machine, MachineError, State, EventData
-from transitions.core import listify, _prep_ordered_arg
-from unittest import TestCase, skipIf
-import warnings
 
 try:
     from unittest.mock import MagicMock
 except ImportError:
-    from mock import MagicMock
+    from mock import MagicMock  # type: ignore
+
+if TYPE_CHECKING:
+    from typing import List, Union, Dict, Callable
 
 
 def on_exit_A(event):
@@ -63,6 +68,20 @@ class TestTransitions(TestCase):
         self.assertEqual(listify((4, 5)), (4, 5))
         self.assertEqual(listify([1, 3]), [1, 3])
 
+        class Foo:
+            pass
+        obj = Foo()
+        proxy = weakref.proxy(obj)
+        del obj
+        self.assertEqual(listify(proxy), [proxy])
+
+    def test_weakproxy_model(self):
+        d = DummyModel()
+        pr = weakref.proxy(d)
+        self.machine_cls(pr, states=['A', 'B'], transitions=[['go', 'A', 'B']], initial='A')
+        pr.go()
+        self.assertTrue(pr.is_B())
+
     def test_property_initial(self):
         states = ['A', 'B', 'C', 'D']
         # Define with list of dictionaries
@@ -85,7 +104,7 @@ class TestTransitions(TestCase):
             {'trigger': 'walk', 'source': 'A', 'dest': 'B'},
             {'trigger': 'run', 'source': 'B', 'dest': 'C'},
             {'trigger': 'sprint', 'source': 'C', 'dest': 'D'}
-        ]
+        ]  # type: List[Union[List[str], Dict[str, str]]]
         m = Machine(states=states, transitions=transitions, initial='A')
         m.walk()
         self.assertEqual(m.state, 'B')
@@ -203,7 +222,7 @@ class TestTransitions(TestCase):
         self.assertEqual(s.message, 'Hello World!')
         s.reverse()
         self.assertEqual(s.state, 'A')
-        self.assertTrue(s.message.startswith('So long'))
+        self.assertTrue(s.message is not None and s.message.startswith('So long'))
 
     def test_before_after_callback_addition(self):
         m = Machine(Stuff(), states=['A', 'B', 'C'], initial='A')
@@ -266,8 +285,8 @@ class TestTransitions(TestCase):
         s.go_f()
         self.assertEqual(s.state, 'F')
         self.assertEqual(s.exit_message, 'E go home...')
-        assert 'I am F!' in s.message
-        assert 'Hello F!' in s.message
+        self.assertIn('I am F!', s.message or "")
+        self.assertIn('Hello F!', s.message or "")
 
     def test_inheritance(self):
         states = ['A', 'B', 'C', 'D', 'E']
@@ -302,16 +321,16 @@ class TestTransitions(TestCase):
         m.add_transition(
             trigger='advance', source='A', dest='B', before='set_message')
         s.advance(message='Hallo. My name is Inigo Montoya.')
-        self.assertTrue(s.message.startswith('Hallo.'))
+        self.assertTrue(s.message is not None and s.message.startswith('Hallo.'))
         s.to_A()
         s.advance('Test as positional argument')
-        self.assertTrue(s.message.startswith('Test as'))
+        self.assertTrue(s.message is not None and s.message.startswith('Test as'))
         # Now wrap arguments in an EventData instance
         m.send_event = True
         m.add_transition(
             trigger='advance', source='B', dest='C', before='extract_message')
         s.advance(message='You killed my father. Prepare to die.')
-        self.assertTrue(s.message.startswith('You'))
+        self.assertTrue(s.message is not None and s.message.startswith('You'))
 
     def test_send_event_data_conditions(self):
         states = ['A', 'B', 'C', 'D']
@@ -332,8 +351,8 @@ class TestTransitions(TestCase):
         self.assertEqual(s.state, 'B')
 
     def test_auto_transitions(self):
-        states = ['A', {'name': 'B'}, State(name='C')]
-        m = Machine('self', states, initial='A', auto_transitions=True)
+        states = ['A', {'name': 'B'}, State(name='C')]  # type: List[Union[str, Dict[str, str], State]]
+        m = Machine(states=states, initial='A', auto_transitions=True)
         m.to_B()
         self.assertEqual(m.state, 'B')
         m.to_C()
@@ -341,13 +360,13 @@ class TestTransitions(TestCase):
         m.to_A()
         self.assertEqual(m.state, 'A')
         # Should fail if auto transitions is off...
-        m = Machine('self', states, initial='A', auto_transitions=False)
+        m = Machine(states=states, initial='A', auto_transitions=False)
         with self.assertRaises(AttributeError):
             m.to_C()
 
     def test_ordered_transitions(self):
         states = ['beginning', 'middle', 'end']
-        m = Machine('self', states)
+        m = Machine(states=states)
         m.add_ordered_transitions()
         self.assertEqual(m.state, 'initial')
         m.next_state()
@@ -359,21 +378,21 @@ class TestTransitions(TestCase):
         self.assertEqual(m.state, 'initial')
 
         # Include initial state in loop
-        m = Machine('self', states)
+        m = Machine(states=states)
         m.add_ordered_transitions(loop_includes_initial=False)
         m.to_end()
         m.next_state()
         self.assertEqual(m.state, 'beginning')
 
         # Do not loop transitions
-        m = Machine('self', states)
+        m = Machine(states=states)
         m.add_ordered_transitions(loop=False)
         m.to_end()
         with self.assertRaises(MachineError):
             m.next_state()
 
         # Test user-determined sequence and trigger name
-        m = Machine('self', states, initial='beginning')
+        m = Machine(states=states, initial='beginning')
         m.add_ordered_transitions(['end', 'beginning'], trigger='advance')
         m.advance()
         self.assertEqual(m.state, 'end')
@@ -381,19 +400,19 @@ class TestTransitions(TestCase):
         self.assertEqual(m.state, 'beginning')
 
         # Via init argument
-        m = Machine('self', states, initial='beginning', ordered_transitions=True)
+        m = Machine(states=states, initial='beginning', ordered_transitions=True)
         m.next_state()
         self.assertEqual(m.state, 'middle')
 
         # Alter initial state
-        m = Machine('self', states, initial='middle', ordered_transitions=True)
+        m = Machine(states=states, initial='middle', ordered_transitions=True)
         m.next_state()
         self.assertEqual(m.state, 'end')
         m.next_state()
         self.assertEqual(m.state, 'beginning')
 
         # Partial state machine without the initial state
-        m = Machine('self', states, initial='beginning')
+        m = Machine(states=states, initial='beginning')
         m.add_ordered_transitions(['middle', 'end'])
         self.assertEqual(m.state, 'beginning')
         with self.assertRaises(MachineError):
@@ -418,17 +437,17 @@ class TestTransitions(TestCase):
         transitions = [['a_to_b', 'A', 'B']]
         # Exception is triggered by default
         b_state = State('B')
-        m1 = Machine('self', states=[a_state, b_state], transitions=transitions,
+        m1 = Machine(states=[a_state, b_state], transitions=transitions,
                      initial='B')
         with self.assertRaises(MachineError):
             m1.a_to_b()
         # Set default value on machine level
-        m2 = Machine('self', states=[a_state, b_state], transitions=transitions,
+        m2 = Machine(states=[a_state, b_state], transitions=transitions,
                      initial='B', ignore_invalid_triggers=True)
         m2.a_to_b()
         # Exception is suppressed, so this passes
         b_state = State('B', ignore_invalid_triggers=True)
-        m3 = Machine('self', states=[a_state, b_state], transitions=transitions,
+        m3 = Machine(states=[a_state, b_state], transitions=transitions,
                      initial='B')
         m3.a_to_b()
         # Set for some states but not others
@@ -440,7 +459,7 @@ class TestTransitions(TestCase):
         with self.assertRaises(MachineError):
             m1.a_to_b()
         # State value overrides machine behaviour
-        m3 = Machine('self', states=[a_state, b_state], transitions=transitions,
+        m3 = Machine(states=[a_state, b_state], transitions=transitions,
                      initial='B', ignore_invalid_triggers=False)
         m3.a_to_b()
 
@@ -468,14 +487,15 @@ class TestTransitions(TestCase):
         before_state_change = MagicMock()
         after_state_change = MagicMock()
 
-        m = Machine('self', states=['A', 'B'],
+        m = Machine(states=['A', 'B'],
                     before_state_change=before_state_change,
                     after_state_change=after_state_change, send_event=True,
                     initial='A', auto_transitions=True)
-
+        self.assertEqual(before_state_change, m.before_state_change[0])
+        self.assertEqual(after_state_change, m.after_state_change[0])
         m.to_B()
-        self.assertTrue(m.before_state_change[0].called)
-        self.assertTrue(m.after_state_change[0].called)
+        self.assertTrue(before_state_change.called)
+        self.assertTrue(after_state_change.called)
 
     def test_state_callbacks(self):
 
@@ -607,11 +627,9 @@ class TestTransitions(TestCase):
         def after_change(machine):
             machine.to_C(machine)
 
-        def failed_transition(machine):
-            raise ValueError('Something was wrong')
-
         states = ['A', 'B', 'C']
-        transitions = [{'trigger': 'do', 'source': '*', 'dest': 'C', 'before': failed_transition}]
+        transitions = [{'trigger': 'do', 'source': '*', 'dest': 'C',
+                        'before': partial(self.stuff.this_raises, ValueError)}]
         m = Machine(states=states, transitions=transitions, queued=True,
                     before_state_change=before_change, after_state_change=after_change)
         with self.assertRaises(MachineError):
@@ -801,7 +819,7 @@ class TestTransitions(TestCase):
         m = Machine(model=[s1, s2], states=states,
                     initial=states[0])
         self.assertEqual(len(m.models), 2)
-        self.assertEqual(len(m.model), 2)
+        self.assertTrue(isinstance(m.model, list) and len(m.model) == 2)
         m.add_transition('advance', 'A', 'B')
         s1.advance()
         self.assertEqual(s1.state, 'B')
@@ -822,6 +840,12 @@ class TestTransitions(TestCase):
         self.assertEqual(s1.state, 'B')
         self.assertEqual(s2.state, 'C')
 
+    def test_remove_model(self):
+        m = self.machine_cls()
+        self.assertIn(m, m.models)
+        m.remove_model(m)
+        self.assertNotIn(m, m.models)
+
     def test_string_trigger(self):
         def return_value(value):
             return value
@@ -845,11 +869,8 @@ class TestTransitions(TestCase):
         model = Model()
         m = Machine(model=model)
         self.assertEqual(model.trigger(5), 5)
-
-        def raise_key_error():
-            raise KeyError
-
-        self.stuff.machine.add_transition('do_raise_keyerror', '*', 'C', before=raise_key_error)
+        self.stuff.machine.add_transition('do_raise_keyerror', '*', 'C',
+                                          before=partial(self.stuff.this_raises, KeyError))
         with self.assertRaises(KeyError):
             self.stuff.trigger('do_raise_keyerror')
 
@@ -870,7 +891,11 @@ class TestTransitions(TestCase):
         self.assertEqual(len(machine.get_triggers('B')), 0)
         self.assertEqual(len(machine.get_triggers('C')), 1)
         # self stuff machine should have to-transitions to every state
-        self.assertEqual(len(self.stuff.machine.get_triggers('B')), len(self.stuff.machine.states))
+        m = self.stuff.machine
+        self.assertEqual(len(m.get_triggers('B')), len(m.states))
+        trigger_name = m.get_triggers('B')
+        trigger_state = m.get_triggers(m.states['B'])
+        self.assertEqual(trigger_name, trigger_state)
 
     def test_skip_override(self):
         local_mock = MagicMock()
@@ -960,13 +985,11 @@ class TestTransitions(TestCase):
         def always_fails(event_data):
             return False
 
-        def always_raises(event_data):
-            raise Exception()
-
         transitions = [
             {'trigger': 'go', 'source': 'A', 'dest': 'B'},
             {'trigger': 'planA', 'source': 'B', 'dest': 'A', 'conditions': always_fails},
-            {'trigger': 'planB', 'source': 'B', 'dest': 'A', 'conditions': always_raises}
+            {'trigger': 'planB', 'source': 'B', 'dest': 'A',
+             'conditions': partial(self.stuff.this_raises, RuntimeError)}
         ]
         m = self.stuff.machine_cls(states=['A', 'B'], transitions=transitions,
                                    finalize_event=finalize_mock, initial='A', send_event=True)
@@ -979,22 +1002,23 @@ class TestTransitions(TestCase):
         self.assertIsInstance(event_data, EventData)
         self.assertEqual(finalize_mock.call_count, 2)
         self.assertFalse(event_data.result)
-        with self.assertRaises(Exception):
+        with self.assertRaises(RuntimeError):
             m.planB()
-        self.assertEqual(finalize_mock.call_count, 3)
+
+        m.finalize_event.append(partial(self.stuff.this_raises, ValueError))
+        # ValueError in finalize should be suppressed
+        # but mock should have been called anyway
+        with self.assertRaises(RuntimeError):
+            m.planB()
+        self.assertEqual(4, finalize_mock.call_count)
 
     def test_machine_finalize_exception(self):
 
-        exception = ZeroDivisionError()
-
-        def always_raises(event):
-            raise exception
-
         def finalize_callback(event):
-            self.assertEqual(event.error, exception)
+            self.assertIsInstance(event.error, ZeroDivisionError)
 
         m = self.stuff.machine_cls(states=['A', 'B'], send_event=True, initial='A',
-                                   before_state_change=always_raises,
+                                   before_state_change=partial(self.stuff.this_raises, ZeroDivisionError),
                                    finalize_event=finalize_callback)
 
         with self.assertRaises(ZeroDivisionError):
@@ -1004,7 +1028,8 @@ class TestTransitions(TestCase):
         self.assertTrue(len(_prep_ordered_arg(3, None)) == 3)
         self.assertTrue(all(a is None for a in _prep_ordered_arg(3, None)))
         with self.assertRaises(ValueError):
-            _prep_ordered_arg(3, [None, None])
+            # deliberately passing wrong arguments
+            _prep_ordered_arg(3, [None, None])  # type: ignore
 
     def test_ordered_transition_callback(self):
         class Model:
@@ -1044,7 +1069,7 @@ class TestTransitions(TestCase):
 
     def test_get_transitions(self):
         states = ['A', 'B', 'C', 'D']
-        m = self.machine_cls('self', states, initial='A', auto_transitions=False)
+        m = self.machine_cls(states=states, initial='A', auto_transitions=False)
         m.add_transition('go', ['A', 'B', 'C'], 'D')
         m.add_transition('run', 'A', 'D')
         self.assertEqual(
@@ -1113,7 +1138,7 @@ class TestTransitions(TestCase):
                 self.status = None
                 self.state = 'some_value'
 
-        m = Machine(Model(), states=['A', 'B'], initial='A', model_attribute='status')
+        m = self.machine_cls(Model(), states=['A', 'B'], initial='A', model_attribute='status')
         self.assertEqual(m.model.status, 'A')
         self.assertEqual(m.model.state, 'some_value')
 
@@ -1165,28 +1190,102 @@ class TestTransitions(TestCase):
         with self.assertRaises(ValueError):
             m.add_transition(m.model_attribute, "A", "B")
 
+    def test_new_state_in_enter_callback(self):
 
-class TestWarnings(TestCase):
-    def test_multiple_machines_per_model(self):
-        class Model:
-            def __init__(self):
-                self.car_state = None
-                self.driver_state = None
+        machine = self.machine_cls(states=['A', 'B'], initial='A')
 
-        instance = Model()
+        def on_enter_B():
+            state = self.machine_cls.state_cls(name='C')
+            machine.add_state(state)
+            machine.to_C()
 
-        with warnings.catch_warnings(record=True) as w:
-            warnings.filterwarnings(action='default', message=r".*transitions version.*", category=DeprecationWarning)
+        machine.on_enter_B(on_enter_B)
+        machine.to_B()
 
-            machine_a = Machine(instance, states=['A', 'B'], initial='A', model_attribute='car_state')
-            machine_b = Machine(instance, states=['A', 'B'], initial='B', model_attribute='driver_state')
-            self.assertEqual(0, len(w))
-            self.assertTrue(instance.is_A())
-            self.assertTrue(instance.is_A())
-            self.assertEqual(1, len(w))
-            self.assertEqual(w[0].category, DeprecationWarning)
-            instance.to_B()
-            self.assertEqual('B', instance.car_state)
-            self.assertFalse(instance.is_A())
-            self.assertEqual(2, len(w))
-            self.assertEqual(w[1].category, DeprecationWarning)
+    def test_on_exception_callback(self):
+        mock = MagicMock()
+
+        def on_exception(event_data):
+            self.assertIsInstance(event_data.error, (ValueError, MachineError))
+            mock()
+
+        m = self.machine_cls(states=['A', 'B'], initial='A', transitions=[['go', 'A', 'B']], send_event=True,
+                             after_state_change=partial(self.stuff.this_raises, ValueError))
+        with self.assertRaises(ValueError):
+            m.to_B()
+        self.assertTrue(m.is_B())
+        with self.assertRaises(MachineError):
+            m.go()
+
+        m.on_exception.append(on_exception)
+        m.to_B()
+        m.go()
+        self.assertTrue(mock.called)
+        self.assertEqual(2, mock.call_count)
+
+    def test_may_transition(self):
+        states = ['A', 'B', 'C']
+        d = DummyModel()
+        m = Machine(model=d, states=states, initial='A', auto_transitions=False)
+        m.add_transition('walk', 'A', 'B')
+        m.add_transition('stop', 'B', 'C')
+        assert d.may_walk()
+        assert not d.may_stop()
+
+        d.walk()
+        assert not d.may_walk()
+        assert d.may_stop()
+
+    def test_may_transition_for_autogenerated_triggers(self):
+        states = ['A', 'B', 'C']
+        m = Machine(states=states, initial='A')
+        assert m.may_to_A()
+        m.to_A()
+        assert m.to_B()
+        m.to_B()
+        assert m.may_to_C()
+        m.to_C()
+
+    def test_may_transition_with_conditions(self):
+        states = ['A', 'B', 'C']
+        d = DummyModel()
+        m = Machine(model=d, states=states, initial='A', auto_transitions=False)
+        m.add_transition('walk', 'A', 'B', conditions=[lambda: False])
+        m.add_transition('stop', 'B', 'C')
+        m.add_transition('run', 'A', 'C')
+        assert not d.may_walk()
+        assert not d.may_stop()
+        assert d.may_run()
+        d.run()
+        assert not d.may_run()
+
+    def test_may_transition_with_auto_transitions(self):
+        states = ['A', 'B', 'C']
+        d = DummyModel()
+        self.machine_cls(model=d, states=states, initial='A')
+        assert d.may_to_A()
+        assert d.may_to_B()
+        assert d.may_to_C()
+
+    def test_machine_may_transitions(self):
+        states = ['A', 'B', 'C']
+        m = self.machine_cls(states=states, initial='A', auto_transitions=False)
+        m.add_transition('walk', 'A', 'B', conditions=[lambda: False])
+        m.add_transition('stop', 'B', 'C')
+        m.add_transition('run', 'A', 'C')
+        m.add_transition('reset', 'C', 'A')
+
+        assert not m.may_walk()
+        assert not m.may_stop()
+        assert m.may_run()
+        m.run()
+        assert not m.may_run()
+        assert not m.may_stop()
+        assert not m.may_walk()
+
+    def test_may_transition_with_invalid_state(self):
+        states = ['A', 'B', 'C']
+        d = DummyModel()
+        m = self.machine_cls(model=d, states=states, initial='A', auto_transitions=False)
+        m.add_transition('walk', 'A', 'UNKNOWN')
+        assert not d.may_walk()
