@@ -4,9 +4,11 @@ except ImportError:
     pass
 
 from .utils import Stuff, DummyModel
-from .test_core import TestTransitions
+from .test_core import TestTransitions, TYPE_CHECKING
 
-from transitions.extensions import MachineFactory
+from transitions.extensions import (
+    LockedGraphMachine, GraphMachine, HierarchicalGraphMachine, LockedHierarchicalGraphMachine
+)
 from transitions.extensions.nesting import NestedState
 from transitions.extensions.states import add_state_features, Timeout, Tags
 from unittest import skipIf
@@ -21,10 +23,14 @@ except ImportError:  # pragma: no cover
     pgv = None
 
 
+if TYPE_CHECKING:
+    from typing import Type, List, Collection, Union
+
+
 @skipIf(pgv is None, 'Graph diagram test requires graphviz.')
 class TestDiagrams(TestTransitions):
 
-    machine_cls = MachineFactory.get_predefined(graph=True)
+    machine_cls = GraphMachine  # type: Type[GraphMachine]
     use_pygraphviz = False
 
     def parse_dot(self, graph):
@@ -51,7 +57,7 @@ class TestDiagrams(TestTransitions):
 
     def setUp(self):
         self.stuff = Stuff(machine_cls=self.machine_cls, extra_kwargs={'use_pygraphviz': self.use_pygraphviz})
-        self.states = ['A', 'B', 'C', 'D']
+        self.states = ['A', 'B', 'C', 'D']  # type: List[Union[str, Collection[str]]]
         self.transitions = [
             {'trigger': 'walk', 'source': 'A', 'dest': 'B'},
             {'trigger': 'run', 'source': 'B', 'dest': 'C'},
@@ -74,7 +80,8 @@ class TestDiagrams(TestTransitions):
 
         for e in edges:
             # label should be equivalent to the event name
-            self.assertIsNotNone(getattr(m, re.match(r'\[label=([^\]]+)\]', e).group(1)))
+            match = re.match(r'\[label=([^\]]+)\]', e)
+            self.assertIsNotNone(match and getattr(m, match.group(1)))
 
         # write diagram to temp file
         target = tempfile.NamedTemporaryFile(suffix='.png', delete=False)
@@ -185,8 +192,8 @@ class TestDiagrams(TestTransitions):
         m.add_transition('to_state_F', 'B', 'F')
         g1 = m.get_graph(show_roi=True)
         dot, nodes, edges = self.parse_dot(g1)
-        self.assertEqual(len(edges), 0)
-        self.assertIn("label=A", dot)
+        self.assertEqual(0, len(edges))
+        self.assertIn(r'label="A\l"', dot)
         # make sure that generating a graph without ROI has not influence on the later generated graph
         # this has to be checked since graph.custom_style is a class property and is persistent for multiple
         # calls of graph.generate()
@@ -195,7 +202,7 @@ class TestDiagrams(TestTransitions):
         _ = m.get_graph()
         g2 = m.get_graph(show_roi=True)
         dot, _, _ = self.parse_dot(g2)
-        self.assertNotIn("label=A", dot)
+        self.assertNotIn(r'label="A\l"', dot)
         m.to_B()
         g3 = m.get_graph(show_roi=True)
         _, nodes, edges = self.parse_dot(g3)
@@ -205,7 +212,7 @@ class TestDiagrams(TestTransitions):
     def test_state_tags(self):
 
         @add_state_features(Tags, Timeout)
-        class CustomMachine(self.machine_cls):
+        class CustomMachine(self.machine_cls):  # type: ignore
             pass
 
         self.states[0] = {'name': 'A', 'tags': ['new', 'polling'], 'timeout': 5, 'on_enter': 'say_hello',
@@ -216,12 +223,12 @@ class TestDiagrams(TestTransitions):
 
     def test_label_attribute(self):
 
-        class LabelState(self.machine_cls.state_cls):
+        class LabelState(self.machine_cls.state_cls):  # type: ignore
             def __init__(self, *args, **kwargs):
                 self.label = kwargs.pop('label')
                 super(LabelState, self).__init__(*args, **kwargs)
 
-        class CustomMachine(self.machine_cls):
+        class CustomMachine(self.machine_cls):  # type: ignore
             state_cls = LabelState
 
         m = CustomMachine(states=[{'name': 'A', 'label': 'LabelA'},
@@ -229,10 +236,10 @@ class TestDiagrams(TestTransitions):
                           transitions=[{'trigger': 'event', 'source': 'A', 'dest': 'B', 'label': 'LabelEvent'}],
                           initial='A', use_pygraphviz=self.use_pygraphviz)
         dot, _, _ = self.parse_dot(m.get_graph())
-        self.assertIn("label=LabelA", dot)
-        self.assertIn("label=NotLabelA", dot)
+        self.assertIn(r'label="LabelA\l"', dot)
+        self.assertIn(r'label="NotLabelA\l"', dot)
         self.assertIn("label=LabelEvent", dot)
-        self.assertNotIn("label=A", dot)
+        self.assertNotIn(r'label="A\l"', dot)
         self.assertNotIn("label=event", dot)
 
     def test_binary_stream(self):
@@ -246,23 +253,42 @@ class TestDiagrams(TestTransitions):
         self.assertEqual(b2, b1.getvalue())
         b1.close()
 
+    def test_graphviz_fallback(self):
+        try:
+            from unittest import mock  # will raise an ImportError in Python 2.7
+            from transitions.extensions.diagrams_graphviz import Graph
+            from transitions.extensions import diagrams_pygraphviz
+            from importlib import reload
+            with mock.patch.dict('sys.modules', {'pygraphviz': None}):
+                # load and reload diagrams_pygraphviz to make sure
+                # an ImportError is raised for pygraphviz
+                reload(diagrams_pygraphviz)
+                m = self.machine_cls(states=['A', 'B', 'C'], initial='A', use_pygraphviz=True)
+            # make sure to reload after test is done to avoid side effects with other tests
+            reload(diagrams_pygraphviz)
+            print(m.graph_cls, pgv)
+            self.assertTrue(issubclass(m.graph_cls, Graph))
+        except ImportError:
+            pass
+
 
 @skipIf(pgv is None, 'Graph diagram test requires graphviz')
 class TestDiagramsLocked(TestDiagrams):
 
-    machine_cls = MachineFactory.get_predefined(graph=True, locked=True)
+    machine_cls = LockedGraphMachine  # type: Type[LockedGraphMachine]
 
 
 @skipIf(pgv is None, 'NestedGraph diagram test requires graphviz')
 class TestDiagramsNested(TestDiagrams):
 
-    machine_cls = MachineFactory.get_predefined(graph=True, nested=True)
+    machine_cls = HierarchicalGraphMachine \
+        # type: Type[HierarchicalGraphMachine | LockedHierarchicalGraphMachine]
 
     def setUp(self):
         super(TestDiagramsNested, self).setUp()
         self.states = ['A', 'B',
                        {'name': 'C', 'children': [{'name': '1', 'children': ['a', 'b', 'c']},
-                                                  '2', '3']}, 'D']
+                                                  '2', '3']}, 'D']  # type: List[Union[str, Collection[str]]]
         self.transitions = [
             {'trigger': 'walk', 'source': 'A', 'dest': 'B'},     # 1 edge
             {'trigger': 'run', 'source': 'B', 'dest': 'C'},      # + 1 edge
@@ -371,7 +397,7 @@ class TestDiagramsNested(TestDiagrams):
         ]
 
         @add_state_features(Timeout, Tags)
-        class CustomStateMachine(self.machine_cls):
+        class CustomStateMachine(self.machine_cls):  # type: ignore
 
             def is_hot(self):
                 return True
@@ -408,4 +434,4 @@ class TestDiagramsLockedNested(TestDiagramsNested):
 
     def setUp(self):
         super(TestDiagramsLockedNested, self).setUp()
-        self.machine_cls = MachineFactory.get_predefined(graph=True, nested=True, locked=True)
+        self.machine_cls = LockedHierarchicalGraphMachine  # type: Type[LockedHierarchicalGraphMachine]
