@@ -1,4 +1,4 @@
-from transitions import Machine
+from transitions import Machine, MachineError
 from transitions.extensions.states import *
 from transitions.extensions import MachineFactory
 from time import sleep
@@ -9,7 +9,7 @@ from .test_graphviz import TestDiagramsLockedNested
 try:
     from unittest.mock import MagicMock
 except ImportError:
-    from mock import MagicMock
+    from mock import MagicMock  # type: ignore
 
 
 class TestTransitions(TestCase):
@@ -196,7 +196,7 @@ class TestStatesDiagramsLockedNested(TestDiagramsLockedNested):
         machine_cls = MachineFactory.get_predefined(locked=True, nested=True, graph=True)
 
         @add_state_features(Error, Timeout, Volatile)
-        class CustomMachine(machine_cls):
+        class CustomMachine(machine_cls):  # type: ignore
             pass
 
         super(TestStatesDiagramsLockedNested, self).setUp()
