@@ -33,6 +33,10 @@ class Stuff(object):
     def this_fails():
         return False
 
+    @staticmethod
+    def this_raises(exception, *args, **kwargs):
+        raise exception
+
     @staticmethod
     def this_fails_by_default(boolean=False):
         return boolean
@@ -54,7 +58,7 @@ class Stuff(object):
         self.message = "Nice to meet you"
 
     def hello_F(self):
-        if not hasattr(self, 'message'):
+        if self.message is None:
             self.message = ''
         self.message += "Hello F!"
 
