@@ -1,15 +1,30 @@
-import itertools
+"""
+    transitions.extensions.asyncio
+    ------------------------------
+
+    This module contains machine, state and event implementations for asynchronous callback processing.
+    `AsyncMachine` and `HierarchicalAsyncMachine` use `asyncio` for concurrency. The extension `transitions-anyio`
+    found at https://github.com/pytransitions/transitions-anyio illustrates how they can be extended to
+    make use of other concurrency libraries.
+    The module also contains the state mixin `AsyncTimeout` to asynchronously trigger timeout-related callbacks.
+"""
+
+# Overriding base methods of states, transitions and machines with async variants is not considered good practise.
+# However, the alternative would mean to either increase the complexity of the base classes or copy code fragments
+# and thus increase code complexity and reduce maintainability. If you know a better solution, please file an issue.
+# pylint: disable=invalid-overridden-method
+
 import logging
 import asyncio
 import contextvars
 import inspect
-from collections import defaultdict, deque
+from collections import deque
 from functools import partial, reduce
 import copy
 
 from ..core import State, Condition, Transition, EventData, listify
 from ..core import Event, MachineError, Machine
-from .nesting import HierarchicalMachine, NestedState, NestedEvent, NestedTransition, _resolve_order
+from .nesting import HierarchicalMachine, NestedState, NestedEvent, NestedTransition, resolve_order
 
 
 _LOGGER = logging.getLogger(__name__)
@@ -17,50 +32,43 @@ _LOGGER.addHandler(logging.NullHandler())
 
 
 class AsyncState(State):
-    """A persistent representation of a state managed by a ``Machine``. Callback execution is done asynchronously.
-
-    Attributes:
-        name (str): State name which is also assigned to the model(s).
-        on_enter (list): Callbacks awaited when a state is entered.
-        on_exit (list): Callbacks awaited when a state is entered.
-        ignore_invalid_triggers (bool): Indicates if unhandled/invalid triggers should raise an exception.
-    """
+    """ A persistent representation of a state managed by a ``Machine``. Callback execution is done asynchronously. """
 
     async def enter(self, event_data):
-        """ Triggered when a state is entered. """
+        """ Triggered when a state is entered.
+        Args:
+            event_data: (AsyncEventData): The currently processed event.
+        """
         _LOGGER.debug("%sEntering state %s. Processing callbacks...", event_data.machine.name, self.name)
         await event_data.machine.callbacks(self.on_enter, event_data)
         _LOGGER.info("%sFinished processing state %s enter callbacks.", event_data.machine.name, self.name)
 
     async def exit(self, event_data):
-        """ Triggered when a state is exited. """
+        """ Triggered when a state is exited.
+        Args:
+            event_data: (AsyncEventData): The currently processed event.
+        """
         _LOGGER.debug("%sExiting state %s. Processing callbacks...", event_data.machine.name, self.name)
         await event_data.machine.callbacks(self.on_exit, event_data)
         _LOGGER.info("%sFinished processing state %s exit callbacks.", event_data.machine.name, self.name)
 
 
 class NestedAsyncState(NestedState, AsyncState):
+    """ A state that allows substates. Callback execution is done asynchronously. """
 
-    async def scoped_enter(self, event_data, scope=[]):
-        self._scope = scope
+    async def scoped_enter(self, event_data, scope=None):
+        self._scope = scope or []
         await self.enter(event_data)
         self._scope = []
 
-    async def scoped_exit(self, event_data, scope=[]):
-        self._scope = scope
+    async def scoped_exit(self, event_data, scope=None):
+        self._scope = scope or []
         await self.exit(event_data)
         self._scope = []
 
 
 class AsyncCondition(Condition):
-    """ A helper class to await condition checks in the intended way.
-
-    Attributes:
-        func (callable): The function to call for the condition check
-        target (bool): Indicates the target state--i.e., when True,
-                the condition-checking callback should return True to pass,
-                and when False, the callback should return False to pass.
-    """
+    """ A helper class to await condition checks in the intended way. """
 
     async def check(self, event_data):
         """ Check whether the condition passes.
@@ -79,19 +87,7 @@ class AsyncCondition(Condition):
 
 
 class AsyncTransition(Transition):
-    """ Representation of an asynchronous transition managed by a ``AsyncMachine`` instance.
-
-    Attributes:
-        source (str): Source state of the transition.
-        dest (str): Destination state of the transition.
-        prepare (list): Callbacks executed before conditions checks.
-        conditions (list): Callbacks evaluated to determine if
-            the transition should be executed.
-        before (list): Callbacks executed before the transition is executed
-            but only if condition checks have been successful.
-        after (list): Callbacks executed after the transition is executed
-            but only if condition checks have been successful.
-    """
+    """ Representation of an asynchronous transition managed by a ``AsyncMachine`` instance. """
 
     condition_cls = AsyncCondition
 
@@ -105,7 +101,7 @@ class AsyncTransition(Transition):
     async def execute(self, event_data):
         """ Executes the transition.
         Args:
-            event_data: An instance of class EventData.
+            event_data (EventData): An instance of class EventData.
         Returns: boolean indicating whether or not the transition was
             successfully executed (True if successful, False if not).
         """
@@ -146,7 +142,7 @@ class AsyncTransition(Transition):
 
 
 class NestedAsyncTransition(AsyncTransition, NestedTransition):
-
+    """ Representation of an asynchronous transition managed by a ``HierarchicalMachine`` instance. """
     async def _change_state(self, event_data):
         if hasattr(event_data.machine, "model_graphs"):
             graph = event_data.machine.model_graphs[id(event_data.model)]
@@ -160,10 +156,14 @@ class NestedAsyncTransition(AsyncTransition, NestedTransition):
             await func()
 
 
+class AsyncEventData(EventData):
+    """ A redefinition of the base EventData intended to easy type checking. """
+
+
 class AsyncEvent(Event):
     """ A collection of transitions assigned to the same trigger """
 
-    async def trigger(self, _model, *args, **kwargs):
+    async def trigger(self, model, *args, **kwargs):
         """ Serially execute all transitions that match the current state,
         halting as soon as one successfully completes. Note that `AsyncEvent` triggers must be awaited.
         Args:
@@ -173,35 +173,16 @@ class AsyncEvent(Event):
         Returns: boolean indicating whether or not a transition was
             successfully executed (True if successful, False if not).
         """
-        func = partial(self._trigger, _model, *args, **kwargs)
-        return await self.machine.process_context(func, _model)
-
-    async def _trigger(self, model, *args, **kwargs):
-        state = self.machine.get_state(getattr(model, self.machine.model_attribute))
-        if state.name not in self.transitions:
-            msg = "%sCan't trigger event %s from state %s!" % (self.machine.name, self.name,
-                                                               state.name)
-            ignore = state.ignore_invalid_triggers if state.ignore_invalid_triggers is not None \
-                else self.machine.ignore_invalid_triggers
-            if ignore:
-                _LOGGER.warning(msg)
-                return False
-            else:
-                raise MachineError(msg)
-        event_data = EventData(state, self, self.machine, model, args=args, kwargs=kwargs)
-        return await self._process(event_data)
-
-    async def _process(self, event_data):
-        await self.machine.callbacks(self.machine.prepare_event, event_data)
-        _LOGGER.debug("%sExecuted machine preparation callbacks before conditions.", self.machine.name)
+        func = partial(self._trigger, EventData(None, self, self.machine, model, args=args, kwargs=kwargs))
+        return await self.machine.process_context(func, model)
 
+    async def _trigger(self, event_data):
+        event_data.state = self.machine.get_state(getattr(event_data.model, self.machine.model_attribute))
         try:
-            for trans in self.transitions[event_data.state.name]:
-                event_data.transition = trans
-                if await trans.execute(event_data):
-                    event_data.result = True
-                    break
-        except Exception as err:
+            if self._is_valid_source(event_data.state):
+                await self._process(event_data)
+        except Exception as err:  # pylint: disable=broad-except; Exception will be handled elsewhere
+            _LOGGER.error("%sException was raised while processing the trigger: %s", self.machine.name, err)
             event_data.error = err
             if self.machine.on_exception:
                 await self.machine.callbacks(self.machine.on_exception, event_data)
@@ -212,68 +193,61 @@ class AsyncEvent(Event):
             _LOGGER.debug("%sExecuted machine finalize callbacks", self.machine.name)
         return event_data.result
 
+    async def _process(self, event_data):
+        await self.machine.callbacks(self.machine.prepare_event, event_data)
+        _LOGGER.debug("%sExecuted machine preparation callbacks before conditions.", self.machine.name)
+        for trans in self.transitions[event_data.state.name]:
+            event_data.transition = trans
+            event_data.result = await trans.execute(event_data)
+            if event_data.result:
+                break
+
 
 class NestedAsyncEvent(NestedEvent):
+    """ A collection of transitions assigned to the same trigger.
+    This Event requires a (subclass of) `HierarchicalAsyncMachine`.
+    """
 
-    async def trigger(self, _model, _machine, *args, **kwargs):
+    async def trigger_nested(self, event_data):
         """ Serially execute all transitions that match the current state,
         halting as soon as one successfully completes. NOTE: This should only
         be called by HierarchicalMachine instances.
         Args:
-            _model (object): model object to
-            _machine (HierarchicalMachine): Since NestedEvents can be used in multiple machine instances, this one
-                                            will be used to determine the current state separator.
-            args and kwargs: Optional positional or named arguments that will
-                be passed onto the EventData object, enabling arbitrary state
-                information to be passed on to downstream triggered functions.
+            event_data (AsyncEventData): The currently processed event.
         Returns: boolean indicating whether or not a transition was
             successfully executed (True if successful, False if not).
         """
-        func = partial(self._trigger, _model, _machine, *args, **kwargs)
-        return await _machine.process_context(func, _model)
-
-    async def _trigger(self, _model, _machine, *args, **kwargs):
-        state_tree = _machine._build_state_tree(getattr(_model, _machine.model_attribute), _machine.state_cls.separator)
-        state_tree = reduce(dict.get, _machine.get_global_name(join=False), state_tree)
-        ordered_states = _resolve_order(state_tree)
+        machine = event_data.machine
+        model = event_data.model
+        state_tree = machine.build_state_tree(getattr(model, machine.model_attribute), machine.state_cls.separator)
+        state_tree = reduce(dict.get, machine.get_global_name(join=False), state_tree)
+        ordered_states = resolve_order(state_tree)
         done = set()
-        res = None
+        event_data.event = self
         for state_path in ordered_states:
-            state_name = _machine.state_cls.separator.join(state_path)
+            state_name = machine.state_cls.separator.join(state_path)
             if state_name not in done and state_name in self.transitions:
-                state = _machine.get_state(state_name)
-                event_data = EventData(state, self, _machine, _model, args=args, kwargs=kwargs)
+                event_data.state = machine.get_state(state_name)
                 event_data.source_name = state_name
                 event_data.source_path = copy.copy(state_path)
-                res = await self._process(event_data)
-                if res:
+                await self._process(event_data)
+                if event_data.result:
                     elems = state_path
                     while elems:
-                        done.add(_machine.state_cls.separator.join(elems))
+                        done.add(machine.state_cls.separator.join(elems))
                         elems.pop()
-        return res
+        return event_data.result
 
     async def _process(self, event_data):
         machine = event_data.machine
         await machine.callbacks(event_data.machine.prepare_event, event_data)
         _LOGGER.debug("%sExecuted machine preparation callbacks before conditions.", machine.name)
 
-        try:
-            for trans in self.transitions[event_data.source_name]:
-                event_data.transition = trans
-                if await trans.execute(event_data):
-                    event_data.result = True
-                    break
-        except Exception as err:
-            event_data.error = err
-            if self.machine.on_exception:
-                await self.machine.callbacks(self.machine.on_exception, event_data)
-            else:
-                raise
-        finally:
-            await machine.callbacks(machine.finalize_event, event_data)
-            _LOGGER.debug("%sExecuted machine finalize callbacks", machine.name)
-        return event_data.result
+        for trans in self.transitions[event_data.source_name]:
+            event_data.transition = trans
+            event_data.result = await trans.execute(event_data)
+            if event_data.result:
+                break
 
 
 class AsyncMachine(Machine):
@@ -295,7 +269,7 @@ class AsyncMachine(Machine):
             Callbacks mentioned here will also be called if a transition or condition check raised an error.
         on_exception: A callable called when an event raises an exception. If not set,
             the Exception will be raised instead.
-        queued (bool): Whether transitions in callbacks should be executed immediately (False) or sequentially.
+        queued (bool or str): Whether transitions in callbacks should be executed immediately (False) or sequentially.
         send_event (bool): When True, any arguments passed to trigger methods will be wrapped in an EventData
             object, allowing indirect and encapsulated access to data. When False, all positional and keyword
             arguments will be passed directly to all callback methods.
@@ -314,13 +288,22 @@ class AsyncMachine(Machine):
     protected_tasks = []
     current_context = contextvars.ContextVar('current_context', default=None)
 
-    def __init__(self, *args, **kwargs):
+    def __init__(self, model=Machine.self_literal, states=None, initial='initial', transitions=None,
+                 send_event=False, auto_transitions=True,
+                 ordered_transitions=False, ignore_invalid_triggers=None,
+                 before_state_change=None, after_state_change=None, name=None,
+                 queued=False, prepare_event=None, finalize_event=None, model_attribute='state', on_exception=None,
+                 **kwargs):
         self._transition_queue_dict = {}
-        super().__init__(*args, **kwargs)
+        super().__init__(model=model, states=states, initial=initial, transitions=transitions,
+                         send_event=send_event, auto_transitions=auto_transitions,
+                         ordered_transitions=ordered_transitions, ignore_invalid_triggers=ignore_invalid_triggers,
+                         before_state_change=before_state_change, after_state_change=after_state_change, name=name,
+                         queued=queued, prepare_event=prepare_event, finalize_event=finalize_event,
+                         model_attribute=model_attribute, on_exception=on_exception, **kwargs)
         if self.has_queue is True:
-            self._transition_queue_dict = _DictionaryMock()
             # _DictionaryMock sets and returns ONE internal value and ignores the passed key
-            self._transition_queue_dict[0] = self._transition_queue
+            self._transition_queue_dict = _DictionaryMock(self._transition_queue)
 
     def add_model(self, model, initial=None):
         super().add_model(model, initial)
@@ -328,7 +311,7 @@ class AsyncMachine(Machine):
             for mod in listify(model):
                 self._transition_queue_dict[id(self) if mod is self.self_literal else id(mod)] = deque()
 
-    async def dispatch(self, trigger, *args, **kwargs):  # ToDo: not tested
+    async def dispatch(self, trigger, *args, **kwargs):
         """ Trigger an event on all models assigned to the machine.
         Args:
             trigger (str): Event name
@@ -367,7 +350,7 @@ class AsyncMachine(Machine):
         """
         Executes callables without parameters in parallel and collects their results.
         Args:
-            partials (list): A list of callable functions
+            callables (list): A list of callable functions
 
         Returns:
             list: A list of results. Using asyncio the list will be in the same order as the passed callables.
@@ -376,15 +359,15 @@ class AsyncMachine(Machine):
 
     async def switch_model_context(self, model):
         """
-        This method is called by an `AsyncTransition` when all conditional tests have passed and the transition will happen.
-        This requires already running tasks to be cancelled.
+        This method is called by an `AsyncTransition` when all conditional tests have passed
+        and the transition will happen. This requires already running tasks to be cancelled.
         Args:
             model (object): The currently processed model
         """
         for running_task in self.async_tasks.get(id(model), []):
             if self.current_context.get() == running_task or running_task in self.protected_tasks:
                 continue
-            elif running_task.done() is False:
+            if running_task.done() is False:
                 _LOGGER.debug("Cancel running tasks...")
                 running_task.cancel()
 
@@ -393,7 +376,7 @@ class AsyncMachine(Machine):
         This function is called by an `AsyncEvent` to make callbacks processed in Event._trigger cancellable.
         Using asyncio this will result in a try-catch block catching CancelledEvents.
         Args:
-            func (callable): The partial of Event._trigger with all parameters already assigned
+            func (partial): The partial of Event._trigger with all parameters already assigned
             model (object): The currently processed model
 
         Returns:
@@ -406,7 +389,7 @@ class AsyncMachine(Machine):
             else:
                 self.async_tasks[id(model)] = [asyncio.current_task()]
             try:
-                res = await self._process(func, model)
+                res = await self._process_async(func, model)
             except asyncio.CancelledError:
                 res = False
             finally:
@@ -414,7 +397,7 @@ class AsyncMachine(Machine):
                 if len(self.async_tasks[id(model)]) == 0:
                     del self.async_tasks[id(model)]
         else:
-            res = await self._process(func, model)
+            res = await self._process_async(func, model)
         return res
 
     def remove_model(self, model):
@@ -431,18 +414,38 @@ class AsyncMachine(Machine):
                 self.models.remove(mod)
         if len(self._transition_queue) > 0:
             queue = self._transition_queue
-            new_queue = [queue.popleft()] + [e for e in queue if e.args[0] not in models]
+            new_queue = [queue.popleft()] + [e for e in queue if e.args[0].model not in models]
             self._transition_queue.clear()
             self._transition_queue.extend(new_queue)
 
-    async def _process(self, trigger, model):
+    async def _can_trigger(self, model, trigger, *args, **kwargs):
+        evt = AsyncEventData(None, None, self, model, args, kwargs)
+        state = self.get_model_state(model).name
+
+        for trigger_name in self.get_triggers(state):
+            if trigger_name != trigger:
+                continue
+            for transition in self.events[trigger_name].transitions[state]:
+                try:
+                    _ = self.get_state(transition.dest)
+                except ValueError:
+                    continue
+                await self.callbacks(self.prepare_event, evt)
+                await self.callbacks(transition.prepare, evt)
+                if all(await self.await_all([partial(c.check, evt) for c in transition.conditions])):
+                    return True
+        return False
+
+    def _process(self, trigger):
+        raise RuntimeError("AsyncMachine should not call `Machine._process`. Use `Machine._process_async` instead.")
+
+    async def _process_async(self, trigger, model):
         # default processing
         if not self.has_queue:
             if not self._transition_queue:
                 # if trigger raises an Error, it has to be handled by the Machine.process caller
                 return await trigger()
-            else:
-                raise MachineError("Attempt to process events synchronously while transition queue is not empty!")
+            raise MachineError("Attempt to process events synchronously while transition queue is not empty!")
 
         self._transition_queue_dict[id(model)].append(trigger)
         # another entry in the queue implies a running transition; skip immediate execution
@@ -464,19 +467,23 @@ class AsyncMachine(Machine):
 
 
 class HierarchicalAsyncMachine(HierarchicalMachine, AsyncMachine):
+    """ Asynchronous variant of transitions.extensions.nesting.HierarchicalMachine.
+        An asynchronous hierarchical machine REQUIRES AsyncNestedStates, AsyncNestedEvent and AsyncNestedTransitions
+        (or any subclass of it) to operate.
+    """
 
     state_cls = NestedAsyncState
     transition_cls = NestedAsyncTransition
     event_cls = NestedAsyncEvent
 
-    async def trigger_event(self, _model, _trigger, *args, **kwargs):
+    async def trigger_event(self, model, trigger, *args, **kwargs):
         """ Processes events recursively and forwards arguments if suitable events are found.
         This function is usually bound to models with model and trigger arguments already
         resolved as a partial. Execution will halt when a nested transition has been executed
         successfully.
         Args:
-            _model (object): targeted model
-            _trigger (str): event name
+            model (object): targeted model
+            trigger (str): event name
             *args: positional parameters passed to the event and its callbacks
             **kwargs: keyword arguments passed to the event and its callbacks
         Returns:
@@ -486,26 +493,78 @@ class HierarchicalAsyncMachine(HierarchicalMachine, AsyncMachine):
                           is not True. Note that a transition which is not executed due to conditions
                           is still considered valid.
         """
-        with self():
-            res = await self._trigger_event(_model, _trigger, None, *args, **kwargs)
-        return self._check_event_result(res, _model, _trigger)
+        event_data = AsyncEventData(state=None, event=None, machine=self, model=model, args=args, kwargs=kwargs)
+        event_data.result = None
 
-    async def _trigger_event(self, _model, _trigger, _state_tree, *args, **kwargs):
+        return await self.process_context(partial(self._trigger_event, event_data, trigger), model)
+
+    async def _trigger_event(self, event_data, trigger):
+        try:
+            with self():
+                res = await self._trigger_event_nested(event_data, trigger, None)
+            event_data.result = self._check_event_result(res, event_data.model, trigger)
+        except Exception as err:  # pylint: disable=broad-except; Exception will be handled elsewhere
+            event_data.error = err
+            if self.on_exception:
+                await self.callbacks(self.on_exception, event_data)
+            else:
+                raise
+        finally:
+            try:
+                await self.callbacks(self.finalize_event, event_data)
+                _LOGGER.debug("%sExecuted machine finalize callbacks", self.name)
+            except Exception as err:  # pylint: disable=broad-except; Exception will be handled elsewhere
+                _LOGGER.error("%sWhile executing finalize callbacks a %s occurred: %s.",
+                              self.name,
+                              type(err).__name__,
+                              str(err))
+        return event_data.result
+
+    async def _trigger_event_nested(self, event_data, _trigger, _state_tree):
+        model = event_data.model
         if _state_tree is None:
-            _state_tree = self._build_state_tree(listify(getattr(_model, self.model_attribute)),
-                                                 self.state_cls.separator)
+            _state_tree = self.build_state_tree(listify(getattr(model, self.model_attribute)),
+                                                self.state_cls.separator)
         res = {}
         for key, value in _state_tree.items():
             if value:
                 with self(key):
-                    tmp = await self._trigger_event(_model, _trigger, value, *args, **kwargs)
+                    tmp = await self._trigger_event_nested(event_data, _trigger, value)
                     if tmp is not None:
                         res[key] = tmp
             if not res.get(key, None) and _trigger in self.events:
-                tmp = await self.events[_trigger].trigger(_model, self, *args, **kwargs)
+                tmp = await self.events[_trigger].trigger_nested(event_data)
                 if tmp is not None:
                     res[key] = tmp
-        return None if not res or all([v is None for v in res.values()]) else any(res.values())
+        return None if not res or all(v is None for v in res.values()) else any(res.values())
+
+    async def _can_trigger(self, model, trigger, *args, **kwargs):
+        state_tree = self.build_state_tree(getattr(model, self.model_attribute), self.state_cls.separator)
+        ordered_states = resolve_order(state_tree)
+        for state_path in ordered_states:
+            with self():
+                return await self._can_trigger_nested(model, trigger, state_path, *args, **kwargs)
+
+    async def _can_trigger_nested(self, model, trigger, path, *args, **kwargs):
+        evt = AsyncEventData(None, None, self, model, args, kwargs)
+        if trigger in self.events:
+            source_path = copy.copy(path)
+            while source_path:
+                state_name = self.state_cls.separator.join(source_path)
+                for transition in self.events[trigger].transitions.get(state_name, []):
+                    try:
+                        _ = self.get_state(transition.dest)
+                    except ValueError:
+                        continue
+                    await self.callbacks(self.prepare_event, evt)
+                    await self.callbacks(transition.prepare, evt)
+                    if all(await self.await_all([partial(c.check, evt) for c in transition.conditions])):
+                        return True
+                source_path.pop(-1)
+        if path:
+            with self(path.pop(0)):
+                return await self._can_trigger_nested(model, trigger, path, *args, **kwargs)
+        return False
 
 
 class AsyncTimeout(AsyncState):
@@ -535,9 +594,9 @@ class AsyncTimeout(AsyncState):
             try:
                 self.on_timeout = kwargs.pop("on_timeout")
             except KeyError:
-                raise AttributeError("Timeout state requires 'on_timeout' when timeout is set.")
+                raise AttributeError("Timeout state requires 'on_timeout' when timeout is set.") from None
         else:
-            self._on_timeout = kwargs.pop("on_timeout", [])
+            self.on_timeout = kwargs.pop("on_timeout", None)
         self.runner = {}
         super().__init__(*args, **kwargs)
 
@@ -556,7 +615,8 @@ class AsyncTimeout(AsyncState):
 
     async def exit(self, event_data):
         """
-        Cancels running timeout tasks stored in `self.runner` first (when not note) before calling further exit callbacks.
+        Cancels running timeout tasks stored in `self.runner` first (when not note) before
+        calling further exit callbacks.
 
         Args:
             event_data (EventData): Data representing the currently processed event.
@@ -607,6 +667,10 @@ class AsyncTimeout(AsyncState):
 
 class _DictionaryMock(dict):
 
+    def __init__(self, item):
+        super().__init__()
+        self._value = item
+
     def __setitem__(self, key, item):
         self._value = item
 
